from typing import Any

import discord

PERMISSIONS = [
	"add_reactions", "administrator", "attach_files", "ban_members", "change_nickname", "connect",
	"create_events", "create_expressions", "create_instant_invite", "create_polls", "create_private_threads",
	"create_public_threads", "deafen_members", "embed_links", "external_emojis", "external_stickers",
	"kick_members", "manage_channels", "manage_emojis", "manage_emojis_and_stickers", "manage_events",
	"manage_expressions", "manage_guild", "manage_messages", "manage_nicknames", "manage_permissions",
	"manage_roles", "manage_threads", "manage_webhooks", "mention_everyone", "moderate_members", "move_members",
	"mute_members", "priority_speaker", "read_message_history", "read_messages", "request_to_speak",
	"send_messages", "send_messages_in_threads", "send_polls", "send_tts_messages", "send_voice_messages",
	"speak", "stream", "use_application_commands", "use_embedded_activities", "use_external_apps",
	"use_external_emojis", "use_external_sounds", "use_external_stickers", "use_soundboard",
	"use_voice_activation", "value", "view_audit_log", "view_channel", "view_creator_monetization_analytics",
	"view_guild_insights"
]


def check_missing_channel_permissions(channel: discord.TextChannel,
                                    permissions: str | list) -> list :
	"""
	Check which permissions are missing for the bot in the specified channel.
	:param channel:
	:param permissions:
	:return:
	"""
	if isinstance(permissions, str) :
		permissions = [permissions]
	bot_perms = channel.permissions_for(channel.guild.me)
	if permissions is None :
		return [perm for perm in PERMISSIONS if not getattr(bot_perms, perm)]
	return [perm for perm in permissions if not getattr(bot_perms, perm)]

def check_missing_guild_permissions(guild: discord.Guild, permissions: str | list[str]) -> None | list[str] :
	"""
	Check if the bot is missing the required permissions in the guild and returns a list with the missing permissions.
	:param guild:
	:param permissions:
	"""
	if isinstance(permissions, str) :
		permissions = [permissions]
	bot_perms = guild.me.guild_permissions
	return [perm for perm in permissions if not getattr(bot_perms, perm)]



def get_bot_permissions(guild: discord.Guild) -> discord.Permissions | dict :
	"""
	Get all the permissions of the bot in the guild.
	:param guild:
	"""
	bot_perms = guild.me.guild_permissions
	permissions = {
		name : getattr(bot_perms, name)
		for name in PERMISSIONS
	}

	return [key for key, value in permissions.items() if value is True]


def check_bot_channel_permissions(channel: discord.TextChannel,
                                        permissions: str | list[str] = None) -> list[str] :
	"""
	Check if the bot has the required permissions in the channel, if no permissions are specified, it will return all the permissions the bot has in the channel.
	:param channel:
	:param permissions:
	:return:

	"""
	if isinstance(permissions, str) :
		permissions = [permissions]
	bot_perms = channel.permissions_for(channel.guild.me)
	if permissions is None :
		return [perm for perm in PERMISSIONS if getattr(bot_perms, perm)]
	return [perm for perm in permissions if getattr(bot_perms, perm)]



import logging

import discord
from discord.ext.commands.help import MISSING

from .exceptions import NoChannelException, NoPermissionException

MAX_LENGTH = 1800


async def send_message(channel: discord.TextChannel | discord.User | discord.Member, message: str = None, embed=None,
                       view=None, files=None,
                       ) -> discord.Message :
	"""Send a message to a channel, if there is no permission it will send an error message to the owner. Automatically handles messages longer than 2000 characters by splitting them into multiple messages."""
	last_message = None
	if channel is None :
		raise NoChannelException
	try :
		length = 0
		if message is None :
			return await channel.send(embed=embed, view=view, files=files)
		while length < len(message) :
			last_message = await channel.send(message[length :length + MAX_LENGTH], embed=embed, view=view, files=files)
			length += MAX_LENGTH
		else :
			return last_message
	except discord.errors.Forbidden :
		required_perms = ['view_channel', 'send_messages', 'embed_links', 'attach_files']
		raise NoPermissionException(required_perms=required_perms, channel=channel)


async def send_response(interaction: discord.Interaction, response, ephemeral=False, view=MISSING, embed=MISSING) :
	"""Sends a response to an interaction. If the interaction is already responded, it will send a followup message."""
	try :

		return await interaction.response.send_message(response, ephemeral=ephemeral, view=view, embed=embed, )
	except discord.errors.Forbidden :
		required_perms = ['view_channel', 'send_messages', 'embed_links', 'attach_files']
		missing_perms = await check_missing_permissions(interaction.channel, required_perms)
		logging.error(f"Missing permission to send message to {interaction.channel.name}")
		await interaction.guild.owner.send(
			f"Missing permission to send message to {interaction.channel.name}. Check permissions: {', '.join(missing_perms)}", )
		raise NoPermissionException(missing_permissions=missing_perms)
	except discord.errors.NotFound or discord.InteractionResponded :
		try :
			await interaction.followup.send(
				response,
				ephemeral=ephemeral,
				view=view,
				embed=embed,

			)
		except discord.errors.NotFound :
			await send_message(interaction.channel, response, view=view, embed=embed)
	except Exception :
		return await interaction.channel.send(response, view=view, embed=embed)


async def await_message(interaction, message) -> discord.Message | bool :
	"""Wait for a message from the user that triggered the interaction. If the message is 'cancel', return False."""
	msg: discord.Message = await send_message(interaction.channel,
	                                          message)

	def check_message(m, interaction) :
		return m.author == interaction.user and m.channel == interaction.channel

	m = await interaction.client.wait_for('message', check=lambda m : check_message(m, interaction), timeout=600)
	await msg.delete()
	if m.content.lower() == "cancel" :
		return False
	return m


async def fetch_message_or_none(channel: discord.TextChannel | discord.DMChannel | discord.Thread,
                                id: int) -> discord.Message | None :
	try :
		return await channel.fetch_message(id)
	except discord.errors.NotFound :
		return None


async def delete_message(message: discord.Message | discord.Thread) :
	try :
		await message.delete()
	except discord.errors.Forbidden :
		if isinstance(message, discord.Message) :
			await send_message(message.channel, f"Could not delete message {message.jump_url}")
			return
		await send_message(message, f"Could not delete message {message.jump_url}")
	except Exception as e :
		logging.error(
			f"Failed to delete {message} because {e}")



import discord

from .permissions import check_missing_channel_permissions, check_missing_guild_permissions


class NoPermissionException(Exception) :
	"""Raised when the bot does not have permission to send a message"""

	def __init__(self, required_perms: str | list, channel: discord.TextChannel = None, guild: discord.Guild = None,
	             message="Missing permission to send message: ", ) :
		if isinstance(channel, discord.TextChannel) :
			self.message = message + ", ".join(check_missing_channel_permissions(channel, required_perms))
			super().__init__(self.message)
		if isinstance(guild, discord.Guild) :
			self.message = message + ", ".join(check_missing_guild_permissions(guild, required_perms))
			super().__init__(self.message)

	def __str__(self) :
		return self.message


class NoChannelException(Exception) :
	"""Raised when the server does not have a channel set to send a message"""

	def __init__(self, message="No channel set or does not exist, check the config or fill in the required arguments.") :
		self.message = message



import logging

import discord

from .exceptions import NoPermissionException
from .messages import send_message


async def ban_member(bans_class, interaction, user, reason, days=1, inform=False) :
	try :
		await bans_class.add_ban(user.id, interaction.guild.id, reason, interaction.user.name)
		await interaction.guild.ban(user, reason=reason, delete_message_days=days)
		embed = discord.Embed(title=f"{user.name} ({user.id}) banned!", description=f"{reason}", color=discord.Color.red())
		embed.set_footer(text=f"Moderator: {interaction.user.name}, was the user informed? {'Yes' if inform else 'No'}")
		await send_message(interaction.channel, embed=embed)
		if not inform :
			return
		try :
			await user.send(f"You have been banned from {interaction.guild.name} for `{reason}`")
		except discord.errors.Forbidden :
			pass

	except discord.Forbidden :
		error = f"Missing permission to ban user {user.name}({user.id}). Check permissions: ban_members or if the bot is higher in the hierarchy than the user."
		logging.error(error)
		await interaction.channel.send(error)
		raise NoPermissionException(required_perms=['ban_members'], guild=interaction.guild)


async def ban_user(interaction: discord.Interaction, user: discord.User, ban_type, reason_modal, ban_class, inform=True,
                   clean=False) :
	if interaction.guild is None :
		await send_message(interaction.channel, "This command can only be used in a server")
		return
	if user is None :
		await send_message(interaction.channel,
		                   "User not found, bot may not be able to fetch user or an invalid id was provided.")
	if user.id == interaction.user.id :
		await send_message(interaction.channel, "You can't ban yourself")
		return
	if user.id == interaction.client.user.id :
		await send_message(interaction.channel, "I can't ban myself")
		return
	if user.id == interaction.guild.owner_id :
		await send_message(interaction.channel, "You can't ban the owner of the server")
		return
	if ban_type == "[silent]" or ban_type == "[hidden]" :
		inform = False

	reason = f"{ban_type}{reason_modal}"

	await ban_member(ban_class, interaction, user, reason, days=1 if clean else 0, inform=inform)

async def dm_user(interaction, reason_modal, user) :
	try :
		await user.send(f"You have been banned from {interaction.guild.name} for `{reason_modal}`")
	except discord.errors.Forbidden :
		pass